/*
 * Copyright 2023, ttcoder
 * All rights reserved. Distributed under the terms of the MIT license.
 */

// genode's base.a
#include <os/reporter.h>  // member

// libc.lib.so
#include <stdio.h>

// haiku.lib.so
#include <app/Roster.h>
#include <interface/Button.h>
#include <interface/StringView.h>
#include <interface/Window.h>
#include <support/StringList.h>
#include <storage/Directory.h>
#include <storage/File.h>



class Win : public BWindow
{
	typedef BWindow inherited;

public:
		Win();
		
		void MessageReceived( BMessage * msg ) override;

private:  // Code
		void lineFeed( BRect & r );
		BButton & add( BRect & r, const char * label, const char * pkg );
		void downloadInstallPkg( const char * pkg );
		void deployRunPkg( const char * pkg );

private:  // Data
		Genode::Reporter installerTx;

private:  // Constants
		enum
		{
			msgbase = 'msg0',
			
			DOWNLOAD,    // AddString( "pkg".. )
			RUN  // AddString( "pkg".. )
		};
};



//#pragma mark -


Win::Win()
:	inherited( BRect(50., 50., 330., 450.), "GeDepot", B_DOCUMENT_WINDOW, B_QUIT_ON_WINDOW_CLOSE | B_ASYNCHRONOUS_CONTROLS )
	,installerTx( be_app->Env(), "installation", "installation_for_manager" )
{
	BRect r( 0., 0., 65., 30. );
	lineFeed( r );
	
	// Create install/run buttons for some selected packages, from these depot/users:
	// - https://depot.genode.org/genodelabs
	// - https://depot.genode.org/cnuke
	// - https://depot.genode.org/cproc
	// - https://depot.genode.org/jschlatow
	// - https://depot.genode.org/skalk
	//
	add( r, "nano3d:", "genodelabs/pkg/nano3d/2023-04-25" );
	add( r, "MESA-cpu:", "genodelabs/pkg/mesa_gpu-cpu/2023-04-25" )
		.Hide();
	add( r, "gears:", "genodelabs/pkg/mesa_gears/2023-10-24" );
	add( r, "textedit:", "genodelabs/pkg/qt5_textedit/2023-10-24" );
///ToDo: find a build of Doom that works in 23.11, or retire this :-)
//	add( r, "doom (23.05)", "cnuke/pkg/chocolate-doom/2023-04-27" );  // no longer works in 23.08+, symbol not found: genode_envp
//	add( r, "chocolate-doom-skalk:", "skalk/pkg/chocolate-doom/2023-10-23" );   NO GO, blocks after opening Audio session to mixer...
	add( r, "falkon:", "cproc/pkg/falkon-jemalloc/2023-10-23" );
	add( r, "morph:", "cproc/pkg/morph_browser/2023-04-26" );
	add( r, "VNC server:", "jschlatow/pkg/vnc_server/2023-11-01" );
}


BButton & Win::add( BRect & r, const char * label, const char * pkg )
{
	BMessage dl( DOWNLOAD );
	BMessage run( RUN );
	BButton * b = nullptr;
	
	AddChild( new BStringView(r, "stringview", label) ); r.OffsetBy( 90., 0. );
	dl.SetString( "pkg", pkg );  // e.g. "genodelabs/pkg/nano3d/2023-04-25"
	run.SetString( "pkg", pkg );
	b = new BButton( r, "button", "Download", new BMessage(dl) );
	AddChild( b ); r.OffsetBy( 80., 0. );
	b = new BButton( r, "button", "Run", new BMessage(run) );
	AddChild( b ); r.OffsetBy( 80., 0. );
	
	lineFeed( r );
	
	return *b;  // so that caller may Hide() it in some (special) cases
}

void Win::lineFeed( BRect & r )
{
	r.OffsetTo( 10., r.bottom +10. );
}

void Win::MessageReceived( BMessage * msg )
{
	switch( msg->what )
	{
		case DOWNLOAD:
			msg->PrintToStream();
			downloadInstallPkg( msg->FindString("pkg") );
		break;
		
		case RUN:
			msg->PrintToStream();
			deployRunPkg( msg->FindString("pkg") );
		break;
		
		default:
			inherited::MessageReceived( msg );
	}
}

void Win::downloadInstallPkg( const char * pkg )
{
	installerTx.enabled( true );
	installerTx.clear();
	
	///ToDo-2: also query index?: <index   path="genodelabs/index/23.04" verify="no" />
	/*
		Or:
			<config arch=\"x86_64\">
				<index user=\"genodelabs\" version=\"19.02\" content=\"yes\"/>
	*/
	///later: set "verify=yes" to check integrity of package (must provide pub key)
	BString s;
	s
		<< "<installation arch=\"x86_64\">\n"
		<< "	<archive path=\"" << pkg << "\"  verify=\"no\"    source=\"no\" />\n"
		<< "</installation>\n"
		;
	
	installerTx.report( s, s.Length() );
}

void Win::deployRunPkg( const char * ge_pkg )
{
	//const char * argv[2] = { pkg, NULL };
	//load_image( 1, argv, NULL );
	//resume_thread..
	
	// BRoster.Launch() only works with actual files, with /boot prefix, so map e.g.
	// - genodelabs/pkg/nano3d/2023-04-25
	// to
	// - /boot/depot/genodelabs/pkg/nano3d/2023-04-25/runtime
	//
	// Downstream, DeployPilot will simply strip those off.
	//
	BString roster_style_pkg;
	{
		roster_style_pkg << "/boot/depot" << "/" << ge_pkg << "/" << "runtime" ;
	}
	
	entry_ref ref;
	
	auto res = get_ref_for_path( roster_style_pkg, &ref );
	printf("get_ref_for_path <%s> -> %d\n", ref.name, res);
	
	printf("entry.exists: %d\n", BEntry(&ref).Exists());
	
	res = be_roster->Launch( &ref );
	printf("launch <%s> -> %d\n", ref.name, res);
}



//#pragma mark -


class App : public BApplication
{
	typedef BApplication inherited;
public:
		App()
		:	inherited( "application/x-vnd.HoG-GeDepot" )
		{
		}
		
		void ReadyToRun() override
		{
			(new Win)->Show();
		}
};


static void create_depot_folders()
{
	status_t res = 0;
	
	res = create_directory( "/boot/public", 0777 );
	printf( "  ..createdir 'public' -> 0x%x\n", res );
	
	res = create_directory( "/boot/depot/genodelabs", 0777 );
	printf( "  ..createdir -> 0x%x\n", res );
	res = create_directory( "/boot/depot/cnuke", 0777 );
	printf( "  ..createdir -> 0x%x\n", res );
	res = create_directory( "/boot/depot/cproc", 0777 );
	printf( "  ..createdir -> 0x%x\n", res );
	res = create_directory( "/boot/depot/jschlatow", 0777 );
	printf( "  ..createdir -> 0x%x\n", res );
	res = create_directory( "/boot/depot/skalk", 0777 );
	printf( "  ..createdir -> 0x%x\n", res );
	
	const char url[] = "https://depot.genode.org";
	size_t written = 0;
	
	written = BFile( "/boot/depot/genodelabs/download", B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE )
		.Write( url, strlen(url) );
	if( written != strlen(url) ) printf( "** failed creating depot 'download' url\n" );
	written = BFile( "/boot/depot/cnuke/download", B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE )
		.Write( url, strlen(url) );
	if( written != strlen(url) ) printf( "** failed creating depot 'download' url\n" );
	written = BFile( "/boot/depot/cproc/download", B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE )
		.Write( url, strlen(url) );
	if( written != strlen(url) ) printf( "** failed creating depot 'download' url\n" );
	written = BFile( "/boot/depot/jschlatow/download", B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE )
		.Write( url, strlen(url) );
	if( written != strlen(url) ) printf( "** failed creating depot 'download' url\n" );
	written = BFile( "/boot/depot/skalk/download", B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE )
		.Write( url, strlen(url) );
	if( written != strlen(url) ) printf( "** failed creating depot 'download' url\n" );
}


int main()
{
	// the <import> tag fails with vfs_ntfs (OPEN_ERR_UNACCESSIBLE...), so let's create the files 'by hand' here instead:
	create_depot_folders();
	
	App().Run();
	
	return 0;
}


